
#include "libvdev/sglib.h"
#include "libvdev/util.h"
#include "libvdev/misc.h"
#include "libvdev/sbuf.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <stdbool.h>
#include <mntent.h>
#include <sys/stat.h>

#include <poll.h>
#include <linux/netlink.h>
#include <sys/socket.h>
#include <fcntl.h>
#include <sys/types.h>
#include <signal.h>

const char *preseed_path = "/lib/vdev/dev-setup.sh";
const char *mountpoint = "/dev";
const char *sysfs_mountpoint = "/sys";
const char *config_path = "/etc/vdev/vdevd.initrd.conf";
const char *lockfile = "/run/vdev/vdev_init_top.lock";

const char *progname = "vdev_init_top";

#define VDEV_LINUX_NETLINK_BUF_MAX 4097
#define VDEV_DEVICE_PATH_UNKNOWN        "UNKNOWN"

static int fd = -1;
static struct flock fl;

typedef char* cstr;

// prototypes
SGLIB_DEFINE_VECTOR_PROTOTYPES (cstr);
SGLIB_DEFINE_VECTOR_FUNCTIONS (cstr);

// connection to the linux kernel for hotplug
struct vdev_linux_context {
        // netlink address
        struct sockaddr_nl nl_addr;
   
        // poll on the netlink socket
        struct pollfd pfd;
   
        // path to mounted sysfs 
        char sysfs_mountpoint[ PATH_MAX+1 ];
};

struct sysfs_scan_context {
        char* uevent_path;
        struct sglib_cstr_vector* frontier;
};

// scan a directory in /sys/devices directory, to find its child directories, that are pushed back to scan_ctx->frontier
// return 0 on success
// return -ENOMEM on OOM
// return -errno on failure to stat
static int scan_device_directory(char const* fp, void* cls) {
        int rc = 0;  
        struct stat sb;
        char *fp_base = NULL;
        char *fp_dup = NULL;
        struct sysfs_scan_context *scan_ctx = (struct sysfs_scan_context*)cls;
        struct sglib_cstr_vector* frontier = scan_ctx->frontier;
   
        fp_base = rindex(fp, '/')+1;
        if (!fp_base)
                return 0;
   
        if (strcmp(fp_base, ".") == 0 || strcmp(fp_base, "..") == 0)
                return 0;
  
        rc = lstat(fp, &sb);
        if (rc != 0) {
                vdev_error ("%s: lstat('%s'): '%s'\n", progname, fp, strerror(errno));
                return -errno;
        }
   
        if (!S_ISDIR(sb.st_mode) && strcmp(fp_base, "uevent") != 0)
                return 0;
   
        fp_dup = vdev_strdup_or_null(fp);
        if (!fp_dup)
                return -ENOMEM;
   
        if (S_ISDIR(sb.st_mode)) {
                rc = sglib_cstr_vector_push_back (frontier, fp_dup);
                if (rc != 0) {
                        free (fp_dup);
                        return rc;
                }
        }
        else {
                // this is a uevent; this directory is a device
                scan_ctx->uevent_path = fp_dup;
        }
   
        return 0;
}

// get a uevent from a uevent file 
// replace newlines with '\0', making the uevent look like it came from the netlink socket
// (i.e. so it can be parsed by vdev_linux_parse_request)
// return 0 on success
// return -ENOMEM on OOM
// return -errno on failure to stat or read
static int vdev_read_uevent(char const* fp_uevent, char** ret_uevent_buf, size_t* ret_uevent_len) {
        int rc = 0;
        struct stat sb;
        char* uevent_buf = NULL;
        size_t uevent_buf_len = 0;
        size_t uevent_len = 0;
   
        // get uevent size  
        rc = stat (fp_uevent, &sb);
        if (rc != 0) {
                vdev_error ("%s: stat('%s'): '%s'\n", progname, fp_uevent, strerror(errno));
                return -errno;
        }
        else
                uevent_buf_len = sb.st_size;
   
        // read the uevent as long as it's a regular file
        if(fp_uevent != NULL && (sb.st_mode & S_IFMT) == S_IFREG) {
                uevent_buf = VDEV_CALLOC (char, uevent_buf_len);
                if (uevent_buf == NULL)
                        return -ENOMEM;
      
                rc = vdev_read_file(fp_uevent, uevent_buf, uevent_buf_len);
                if (rc != 0) {
                        vdev_error ("%s: read_file('%s'): '%s'\n", progname, fp_uevent, strerror (errno));
                        free (uevent_buf);
                }
                else {
                        for(unsigned int i = 0; i < uevent_buf_len; i++) {
                                if (uevent_buf[i] == '\n')               
                                        uevent_buf[i] = '\0';
                        }
         
                        // NOTE: the stat size is an upper-bound.  Find the exact number of bytes.
                        for(uevent_len = 0; uevent_len < uevent_buf_len; ) {
                                if (*(uevent_buf + uevent_len) == '\0')
                                        break;
                                uevent_len += strlen( uevent_buf + uevent_len ) + 1;
                        }
                        *ret_uevent_buf = uevent_buf;
                        *ret_uevent_len = uevent_len;
                }
        }
   
        return rc;
}

// make the full sysfs path from the dev path, plus an additional path 
// return NULL on OOM
static char* vdev_linux_sysfs_fullpath(char const* sysfs_mountpoint, char const* devpath, char const* attr_path) {
        char* tmp = NULL;
        char* ret = NULL;
   
        tmp = vdev_fullpath(sysfs_mountpoint, devpath, NULL);
        if (!tmp)
                return NULL;
   
        ret = vdev_fullpath(tmp, attr_path, NULL);
        free(tmp);
   
        return ret;
}

// read the kernel-given device subsystem from sysfs 
// return 0 on success, and set *subsystem
// return -ENOMEM on OOM 
// return negative on readlink failure
static int vdev_linux_sysfs_read_subsystem(const char *mountpoint, char const* devpath, char** subsystem) {
        int rc = 0;
        char linkpath[PATH_MAX+1];
        size_t linkpath_len = PATH_MAX;
        char* subsystem_path = NULL;
   
        memset(linkpath, 0, PATH_MAX+1);
   
        subsystem_path = vdev_linux_sysfs_fullpath(mountpoint, devpath, "subsystem");
        if (subsystem_path == NULL)
                return -ENOMEM;
   
        rc = readlink(subsystem_path, linkpath, linkpath_len);
        if (rc < 0) {
                //vdev_error("%s: readlink('%s') %s\n", progname, subsystem_path, strerror(errno));
                free(subsystem_path);
                return -errno;
        }
   
        free(subsystem_path);
   
        *subsystem = vdev_basename(linkpath, NULL);
        if (*subsystem == NULL)
                return -ENOMEM;
   
        return 0;
}

static bool symlink_exists(const char *path)
{
    struct stat info;
    return lstat(path, &info) == 0;
}

static int add_device(char const* fp_uevent) {
        int rc = 0;
        struct stat sb;
        char* uevent_buf = NULL;
        size_t uevent_buf_len = 0;
        char majorbuf[10] = {0};
        char minorbuf[10] = {0};
        char devname[PATH_MAX+1];
        char *full_devpath = NULL;
        char *subsystem = NULL;         // sysfs subsystem
        //char *p[] = { "acpi", "leds", "queues", "tty", "wmi", "cpu", "hdaudio", "scsi", "pmi", (char*)0 }; //   
   
        // get uevent
        rc = vdev_read_uevent(fp_uevent, &uevent_buf, &uevent_buf_len);
        if( rc != 0 ) {
                vdev_error("%s: read_uevent('%s'): '%s'\n", progname, fp_uevent, strerror(errno));
                return rc;
        }
   
        if( uevent_buf_len == 0 ) {
                free(uevent_buf);
                return 0;
        }
   
        for (unsigned int i = 0; i < uevent_buf_len;) {
                char *tmp = rindex(uevent_buf + i, '=') + 1;
                if (!strncmp(uevent_buf+i, "DEVNAME=", strlen("DEVNAME=")) && strlen(tmp) > 0)
                        strncpy_t(devname, sizeof(devname), tmp, strlen(tmp));
                else if (!strncmp(uevent_buf+i, "MAJOR=", strlen("MAJOR=")) && strlen(tmp) > 0)
                        strncpy_t(majorbuf, sizeof(majorbuf), tmp, strlen(tmp));
                else if (!strncmp(uevent_buf+i, "MINOR=", strlen("MINOR=")) && strlen(tmp) > 0)
                        strncpy_t(minorbuf, sizeof(minorbuf), tmp, strlen(tmp));
                else if (!strncmp(uevent_buf+i, "MODALIAS=", strlen("MODALIAS=")) && strlen(tmp) > 0) {
                        FILE *pfin = NULL;
                        pid_t pid;
                        int wstatus;
                        sbuf_t cmd;
                        char kernel[64];
                        int index = 0;         
                        bool shall_continue = true;
/*
                        while (tmp[index] != ':')
                                index++;

                        strncpy_t(kernel, sizeof(kernel), tmp, index);

                        for (unsigned i = 0; i < sizeof(p)/sizeof(char*) - 1; i++) {
                                if (!strcmp(kernel, p[i])) {
                                        shall_continue = false;
                                        break;
                                }
                        }   
  */ 
                        if (shall_continue) {
                                sbuf_init(&cmd);
                                sbuf_concat(&cmd, 3, "/sbin/modprobe -b -q -- \"", tmp, "\"");
                                pfin = epopen(cmd.buf, &pid);
                                if(pfin) { 
                                        fclose(pfin);
                                        waitpid(pid, &wstatus, 0);
                                }
                                sbuf_free(&cmd);
                        }
                }
                i += strlen(uevent_buf + i) + 1;
        }

        if (!subsystem) {
                char devpath[PATH_MAX+1];
                strncpy_t(devpath, sizeof(devpath), fp_uevent + strlen("/sys"), strlen(fp_uevent) - strlen("/sys") - strlen("/uevent"));

                // see if we have a subsystem 
                rc = vdev_linux_sysfs_read_subsystem("/sys", devpath, &subsystem);
                if (rc == 0 && strlen(majorbuf) > 0 && strlen(minorbuf) > 0) {
                        if (subsystem && strcasecmp(subsystem, "block") == 0) {
         
				// this is a block 
				// mode_t dev_mode = S_IFBLK;
                                
			}
			else {
                                // this is a char 
				// mode_t dev_mode = S_IFCHR;
                                int i = 0;
                                FILE *fptr = NULL;
                                sbuf_t link_source;
                                sbuf_t link_target;
                                sbuf_t metadata_dir;
                                sbuf_t metadata_links;
                                sbuf_t serialized;
                                
                                sbuf_init(&link_source);        
                                sbuf_init(&link_target);
                                sbuf_concat(&link_source, 3, mountpoint, "/", devname);                                                
                                sbuf_concat(&link_target, 5, mountpoint, "/char/", majorbuf, ":", minorbuf);                                
                                //printf("%s\n", link_source.buf);
                                //printf("%s\n", link_target.buf);
                                if (!symlink_exists(link_target.buf) && symlink(link_source.buf, link_target.buf) == -1) {
                                        vdev_error("%s: symlink('%s'): '%s'\n", progname, link_source.buf, strerror(errno));
                                        sbuf_free(&link_source);
                                        sbuf_free(&link_target);
                                        return -errno;
                                }
                                sbuf_free(&link_source);

                                sbuf_init(&metadata_dir);
                                sbuf_concat(&metadata_dir, 5, mountpoint, "/metadata", mountpoint, "/", devname);
                                rc = vdev_mkdirs(metadata_dir.buf, 0, 0755);
                                if (rc != 0) {
                                        vdev_error("vdev_mkdirs('%s') %s\n", metadata_dir.buf, strerror(errno));
                                        sbuf_free(&metadata_dir);
                                        sbuf_free(&link_target);
                                        return -errno;
                                }
                                
                                sbuf_init(&metadata_links);
                                sbuf_concat(&metadata_links, 2, metadata_dir.buf, "/links");
                                fptr = fopen(metadata_links.buf, "a");
                                if (rc != 0) {
                                        vdev_error("fopen('%s') %s\n", metadata_links.buf, strerror(errno));
                                        sbuf_free(&metadata_links);
                                        sbuf_free(&metadata_dir);
                                        sbuf_free(&link_target);
                                        return -errno;
                                }
                                
                                sbuf_init(&serialized);
                                sbuf_concat(&serialized, 2, mountpoint, "/metadata/sysfs/"); 
                                while (i < strlen(fp_uevent)-strlen("/uevent")) {
                                        if (fp_uevent[i] != '/')
                                                sbuf_addch(&serialized, fp_uevent[i]);
                                        else
                                                sbuf_addstr(&serialized, "\\x2f");
                                        i++;
                                }
                                //fprintf(fptr, serialized.buf);
                                //fprintf(fptr, link_target.buf);
                                fclose(fptr);
                                sbuf_free(&link_target);
                                sbuf_free(&metadata_links);
                                
                                //printf("%s\n", metadata_dir.buf);
                                //printf("%s\n", serialized.buf);
                                if (!symlink_exists(serialized.buf) && symlink(metadata_dir.buf, serialized.buf) == -1) {
                                        vdev_error("%s: symlink('%s'): '%s'\n", progname, serialized.buf, strerror(errno));
                                        sbuf_free(&metadata_dir);
                                        sbuf_free(&serialized);
                                        return -errno;
                                }
                                
                                sbuf_free(&metadata_dir);
                                sbuf_free(&serialized);
			}
                }
                else if (rc != -ENOMEM ) {
                        // this is weird...
                        ///vdev_warn("no subsystem found for '%s'\n", devpath );
                        rc = 0;
                }
        }

        if (subsystem)
                free(subsystem );

        // Try to create a lock file. Bear in mind that the system still may be read only at this point.
        // If so, we ignore the failed atempt because we'll try again when loading the next driver 
        if (fd == -1) {
                fd = open(lockfile, O_RDWR|O_CREAT, 0640); 
                if (fd < 0) {
                        switch (errno) {
                        case ENOMEM:
                                return -ENOMEM;
                        default:
                                return rc;
                        }      
                }
        } else {   
                return rc;
                }   

        memset(&fl, 0, sizeof(fl));
   
        fl.l_type = F_WRLCK; // Set a write lock
        fl.l_whence = SEEK_SET; // Start at beginning of file
        fl.l_start = 0; // Offset from beginning of file
        fl.l_len = 0; // Lock entire file
   
        // F_SETLK(W) ignores it; F_OFD_SETLK(W) requires it to be zero
        fl.l_pid = 0; //getpid(); // Set process ID
   
        // F_SETLKW specifies blocking mode
        fcntl(fd, F_SETLKW, &fl); // Set the lock

        return rc;
}

/**
 * \brief Variadic function
 */
static int find_devices_at_frontier (const char *sysfs_path, const char *device_frontier, struct sglib_cstr_vector *uevent_paths) {
        int rc = 0;
        struct sglib_cstr_vector frontier;
        struct sysfs_scan_context scan_ctx;

        sglib_cstr_vector_init (&frontier);
   
        memset(&scan_ctx, 0, sizeof(struct sysfs_scan_context));
   
        scan_ctx.frontier = &frontier;

        rc = vdev_load_all (device_frontier, scan_device_directory, &scan_ctx);
        if (rc != 0) {
                vdev_error ("%s: vdev_load_all('%s'): '%s'\n", progname, device_frontier, strerror(errno));
                sglib_cstr_vector_free (&frontier);      
                return rc;
        }
   
        while (1) {   
                char* dir = NULL;
                unsigned long len = sglib_cstr_vector_size (&frontier);
                if (len == 0)
                        break;

                dir = sglib_cstr_vector_at (&frontier, len - 1);
                sglib_cstr_vector_set (&frontier, NULL, len - 1);
      
                sglib_cstr_vector_pop_back (&frontier);
      
                // scan for more devices 
                rc = vdev_load_all (dir, scan_device_directory, &scan_ctx);
                if (rc != 0) {
                        vdev_error ("%s: vdev_load_all('%s'): '%s'\n", progname, dir, strerror(errno));
                        free (dir);
                        break;
                }
      
                // is one of them a uevent?
                if (scan_ctx.uevent_path) {
                        bool is_ok = false;
                        char* uevent_path = vdev_strdup_or_null(scan_ctx.uevent_path);
                        if (!uevent_path) {
                                free (dir);
                                free (scan_ctx.uevent_path);
                                scan_ctx.uevent_path = NULL;
                                rc = -ENOMEM;
                                break;
                        }

                        if (sysfs_get_parent_with_subsystem (uevent_path, "pci")) {
                                char *subsystem = NULL;
                                char path[PATH_MAX+1];
                                strncpy_t(path, sizeof(path), uevent_path, strlen(uevent_path)-strlen("/uevent"));
                                vdev_linux_sysfs_read_subsystem("", path, &subsystem);
                                if (subsystem) {
                                        char *p[] = { "pci", "block", (char*)0 };
                                        for (unsigned j = 0; j < sizeof(p)/sizeof(char*) - 1; j++) {
                                                if (!strcmp(subsystem, p[j])) {
                                                        is_ok = true;
                                                        break;
                                                }
                                        }
                                        free(subsystem);
                                }
                        }
                        
                        if (!is_ok) {
                                free (uevent_path);
                                continue;
                        }

                        rc = sglib_cstr_vector_push_back (uevent_paths, uevent_path);
                        if (rc < 0) {
                                free (dir);
                                free (scan_ctx.uevent_path);
                                scan_ctx.uevent_path = NULL;
                                break;
                        }
         
                        free (scan_ctx.uevent_path);
                        scan_ctx.uevent_path = NULL;
         
                        rc = add_device(uevent_path);
                        if (rc != 0)
                                vdev_error("%s: add_device('%s'): '%s'\n", progname, uevent_path, strerror(errno));
                }
                free (dir);
        }
        sglib_cstr_vector_free (&frontier);
        return rc;
}

static int find_devices(const char *sysfs_mountpoint, struct sglib_cstr_vector* uevent_paths) {
        int rc = 0;
        struct sglib_cstr_vector frontier;

        struct sysfs_scan_context scan_ctx;
   
        memset (&scan_ctx, 0, sizeof(struct sysfs_scan_context));
      
        char* devroot = NULL;

        sglib_cstr_vector_init (&frontier);
   
        scan_ctx.frontier = &frontier;
   
        devroot = vdev_fullpath (sysfs_mountpoint, "/devices", NULL);     // devroot <= /sys/devices
        if( devroot == NULL )      
                return -ENOMEM;
    
        rc = vdev_load_all (devroot, scan_device_directory, &scan_ctx);
        if( rc != 0 ) {
                vdev_error ("%s: vdev_load_all('%s'): '%s'\n", progname, devroot, strerror(errno)); 
                free (devroot);    
                sglib_cstr_vector_free (&frontier);      
                return rc;
        }
   
        free (scan_ctx.uevent_path);
        scan_ctx.uevent_path = NULL;
   
        for (unsigned long i = 0; i < sglib_cstr_vector_size (&frontier); i++)
                find_devices_at_frontier (sysfs_mountpoint, sglib_cstr_vector_at (&frontier, i), uevent_paths); 
      
        sglib_cstr_vector_free (&frontier);
   
        return rc;
}

static void vdev_linux_context_free (struct vdev_linux_context *ctx) {
        if( ctx != NULL ) {
                if( ctx->pfd.fd >= 0 ) {
                        close( ctx->pfd.fd );
                        ctx->pfd.fd = -1;
                }
        }
}

int main (int argc, char **argv) {
        int rc, flag;
        struct sglib_cstr_vector uevent_paths;
        char * buf = NULL;
        ssize_t nr = 0;
        struct vdev_linux_context ctx;
        sbuf_t char_dir;
        sbuf_t sysfs_mntp;

        FILE *pfin = NULL;
        pid_t pid;
        int wstatus;
        char* command = NULL;
   
        /* netlink */
        ssize_t len = 0;
        char uevent_buf[VDEV_LINUX_NETLINK_BUF_MAX];
        char cbuf[CMSG_SPACE(sizeof(struct ucred))];
        struct cmsghdr *chdr = NULL;
        struct ucred *cred = NULL;
        struct msghdr hdr;
        struct iovec iov;
        struct sockaddr_nl cnls;
        size_t slen = 128 * 1024 * 1024;
        int so_passcred_enable = 1;
   
        command = VDEV_CALLOC(char, strlen(preseed_path)+4+strlen(mountpoint)+2+strlen(config_path)+1);
        if (!command)
                return -ENOMEM;  // OOM
   
        sprintf(command, "%s 1 %s %s", preseed_path, mountpoint, config_path );
         
        pfin = epopen(command, &pid);
        if (pfin) { 
               fclose(pfin);
               waitpid(pid, &wstatus, 0);
        }
        free( command );
        command = NULL;

        sbuf_init(&char_dir);
        sbuf_concat(&char_dir, 2, mountpoint, "/char");
        rc = vdev_mkdirs(char_dir.buf, 0, 0755);
        if (rc != 0) {
                vdev_error("vdev_mkdirs(%s): %s\n", char_dir.buf, strerror(errno));
                sbuf_free(&char_dir);
                return -errno;
        }
        
        sbuf_init(&sysfs_mntp);
        sbuf_concat(&sysfs_mntp, 2, mountpoint, "/metadata/sysfs");
        rc = vdev_mkdirs(sysfs_mntp.buf, 0, 0755);
        if (rc != 0) {
                vdev_error("vdev_mkdirs(%s): %s\n", sysfs_mntp.buf, strerror(errno));
                sbuf_free(&sysfs_mntp);
                return -errno;
        }      
        sbuf_free(&sysfs_mntp);
        
        if (chdir(char_dir.buf) != 0) {
                vdev_error("chdir(%s): %s\n", char_dir.buf, strerror(errno));
                sbuf_free(&char_dir);
                return -errno;
        }      
        sbuf_free(&char_dir);
        
        send_to_background();
 
        /*  Coldplug phase: get already existent uevents in the sysfs topology  */

        sglib_cstr_vector_init (&uevent_paths);
        rc = find_devices (sysfs_mountpoint, &uevent_paths);
        sglib_cstr_vector_free (&uevent_paths);      
        vdev_linux_context_free (&ctx);
  
        if (fd > 0) {
                fl.l_type = F_UNLCK; // Release the lock
                fcntl (fd, F_SETLK, &fl); // Unlock the file
                close (fd);
                unlink(lockfile);
        }

        return 0;
}
